/*
 * taskview-manager.c
 *
 * TaskviewManager
 *
 * Copyright © 2010 Salomon Sickert
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 3 of the licence or (at
 * your option) any later version.
 *
 * See the included COPYING.LESSER file for more information.
 *
 * Authors:
 *    Salomon Sickert <sickert@in.tum.de>
 */

#include "taskview.h"
#include "taskview-manager.h"
#include "taskview-generic.h"
#include "taskview-io.h"

/* ----------------------------------------------------------------------------------------------------
 * Private Data Declaration
 * ---------------------------------------------------------------------------------------------------- */

#define TASKVIEW_MANAGER_GET_PRIVATE(obj)	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), TASKVIEW_TYPE_MANAGER, TaskviewManagerPrivate))

G_DEFINE_TYPE (TaskviewManager, taskview_manager, G_TYPE_OBJECT);

struct _TaskviewManagerPrivate
{
  GDBusConnection* connection;
  guint owner_id;
  guint registration_id;
  guint registration_id_subtree;
  GHashTable *cache;
};

/* ----------------------------------------------------------------------------------------------------
 * GDBus Interface Specification
 * ---------------------------------------------------------------------------------------------------- */

static GDBusNodeInfo *introspection_data = NULL;
static GDBusInterfaceInfo *interface_info = NULL;

static const gchar introspection_xml[] =
  "<node>"
  "  <interface name=\"org.gnome.TaskView.Manager\">"
  "    <method name=\"TaskViewList\">"
  "      <arg type=\"ao\" name=\"task_view_object_path\" direction=\"out\" />"
  "    </method>"
  "    <signal name=\"TaskViewAdded\">"
  "      <arg type=\"o\" name=\"task_view_object_path\" />"
  "    </signal>"
  "    <signal name=\"TaskViewRemoved\">"
  "      <arg type=\"o\" name=\"task_view_object_path\" />"
  "    </signal>"
  "    <property type=\"s\" name=\"ApplicationName\" access=\"read\" />"
  "  </interface>"
  "</node>";

/* ----------------------------------------------------------------------------------------------------
 * Manager Implementation
 * ---------------------------------------------------------------------------------------------------- */

static void
manager_method_call (GDBusConnection       *connection,
                     const gchar           *sender,
                     const gchar           *object_path,
                     const gchar           *interface_name,
                     const gchar           *method_name,
                     GVariant              *parameters,
                     GDBusMethodInvocation *invocation,
                     gpointer               user_data)
{
  if (g_strcmp0 (method_name, "TaskViewList") == 0)
    {
      TaskviewGeneric *view;
      GList *ptr;
      gchar *object_path;
      GVariantBuilder *builder;
      GVariant *result;

      builder = g_variant_builder_new (G_VARIANT_TYPE("ao"));
      ptr = g_hash_table_get_values (taskview_generic_get_views ());

      for (; ptr != NULL; ptr = ptr->next)
        {
          view = TASKVIEW_GENERIC (ptr->data);
          g_object_get (view, "object-path", &object_path, NULL);
          g_variant_builder_add (builder, "o", object_path);
          g_free (object_path);
        }

      result = g_variant_new ("(ao)", builder);

      g_dbus_method_invocation_return_value (invocation, result);

      g_list_free (ptr);
      g_variant_unref (result);
      g_variant_builder_unref (builder);
    }
  else
    {
      g_critical ("I was asserted, that nobody finds this dragon here.");
    }
}


static GVariant *
manager_get_property (GDBusConnection  *connection,
                      const gchar      *sender,
                      const gchar      *object_path,
                      const gchar      *interface_name,
                      const gchar      *property_name,
                      GError          **error,
                      gpointer          user_data)
{
  GVariant *ret;
  ret = NULL;

  if (g_strcmp0 (property_name, "ApplicationName") == 0)
    {
      ret = g_variant_new_string (g_get_application_name());
    }
  else
    {
      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS, "Unknown Property");
    }

  return ret;
}

const GDBusInterfaceVTable manager_vtable =
{
  manager_method_call,
  manager_get_property,
  NULL
};

/* ----------------------------------------------------------------------------------------------------
 * Subtree Implementation
 * ---------------------------------------------------------------------------------------------------- */

static gchar **
subtree_enumerate (GDBusConnection       *connection,
                   const gchar           *sender,
                   const gchar           *object_path,
                   gpointer               user_data)
{
  TaskviewGeneric *view;
  GPtrArray *p;
  GList *ptr;
  gchar **nodes;
  guint id;

  p = g_ptr_array_new ();
  ptr = g_hash_table_get_values (taskview_generic_get_views ());

  for (; ptr != NULL; ptr = ptr->next)
    {
      view = TASKVIEW_GENERIC (ptr->data);
      g_object_get (view, "id", &id, NULL);
      g_ptr_array_add (p, g_strdup_printf ("%i", id));
    }

  g_ptr_array_add (p, NULL);
  nodes = (gchar **) g_ptr_array_free (p, FALSE);
  g_list_free (ptr);
  return nodes;
}


static GDBusInterfaceInfo **
subtree_introspect (GDBusConnection       *connection,
                    const gchar           *sender,
                    const gchar           *object_path,
                    const gchar           *node,
                    gpointer               user_data)
{
  TaskviewGeneric *view;

  GPtrArray *p;

  //  We don't need a root object, so we just return nothing
  if (node == NULL)
    {
      return NULL;
    }
  else
    {
      p = g_ptr_array_new ();

      // We enumerate only integers, so we can just use atoi...
      view = taskview_generic_get_inst_by_id (atoi (node));

      if (TASKVIEW_IS_GENERIC (view))
        g_ptr_array_add (p, g_dbus_interface_info_ref (TASKVIEW_GENERIC_GET_CLASS (view)->interface));
      if (TASKVIEW_IS_IO (view))
        g_ptr_array_add (p, g_dbus_interface_info_ref (TASKVIEW_IO_GET_CLASS (view)->interface));

      g_ptr_array_add (p, NULL);

      return (GDBusInterfaceInfo **) g_ptr_array_free (p, FALSE);
    }

}

static const GDBusInterfaceVTable *
subtree_dispatch (GDBusConnection             *connection,
                  const gchar                 *sender,
                  const gchar                 *object_path,
                  const gchar                 *interface_name,
                  const gchar                 *node,
                  gpointer                    *out_user_data,
                  gpointer                     user_data)
{
  const GDBusInterfaceVTable *vtable_to_return;
  gpointer user_data_to_return;

  /* It can happen that buggy applications call methods of disappered views.
   * We just ignore them and return NULL. */

  user_data_to_return = taskview_generic_get_inst_by_id (atoi (node));

  if (g_strcmp0 (interface_name, TASKVIEW_GENERIC_INTERFACE_NAME) == 0 &&
      TASKVIEW_IS_GENERIC (user_data_to_return))
    {
      vtable_to_return = (TASKVIEW_GENERIC_GET_CLASS (user_data_to_return))->vtable;
      *out_user_data = user_data_to_return;
    }
  else if (g_strcmp0 (interface_name, TASKVIEW_IO_INTERFACE_NAME) == 0 &&
           TASKVIEW_IS_IO (user_data_to_return))
    {
      vtable_to_return = (TASKVIEW_IO_GET_CLASS (user_data_to_return))->vtable;
      *out_user_data = user_data_to_return;
    }
  else
    {
      vtable_to_return = NULL;
      *out_user_data = NULL;
      g_debug ("Buggy application called methods on non existing objectpaths!");
    }

  return vtable_to_return;
}

const GDBusSubtreeVTable subtree_vtable =
{
  subtree_enumerate,
  subtree_introspect,
  subtree_dispatch
};

/* ----------------------------------------------------------------------------------------------------
 * Signal Handler
 * ---------------------------------------------------------------------------------------------------- */

static void
on_name_acquired (GDBusConnection *connection,
                  const gchar     *name,
                  gpointer         user_data)
{
  g_debug ("Acquired Name");
}

static void
on_name_lost (GDBusConnection *connection,
              const gchar     *name,
              gpointer         user_data)
{
  g_warning ("Lost Name, this should _not_ happen...");
}

static char *
generate_bus_name (TaskviewManager *self)
{
  gchar *unique_bus_name;
  gchar *bus_name;

  unique_bus_name = g_strdup (g_dbus_connection_get_unique_name (self->priv->connection));
  g_strcanon (unique_bus_name, "1234567890", '_');
  bus_name = g_strconcat (TASKVIEW_BUS_PREFIX, g_get_prgname (), ".", unique_bus_name, NULL);

  g_assert (g_dbus_is_name (bus_name));
  g_free (unique_bus_name);

  return bus_name;
}

static void
on_bus_acquired  (GObject *source_object,
                  GAsyncResult *res,
                  gpointer user_data)
{
  TaskviewManager *self = TASKVIEW_MANAGER (user_data);
  GDBusConnection *connection;
  GError *error = NULL;
  gchar *bus_name;

  connection = g_bus_get_finish (res, &error);

  if (error != NULL)
    {
      g_error ("%s", error->message);
      g_error_free (error);
			// TODO: Fallback!
			return;
    }

  g_dbus_connection_set_exit_on_close (connection, FALSE);
  self->priv->connection = connection;

  bus_name = generate_bus_name (self);

  self->priv->owner_id = g_bus_own_name_on_connection (connection,
                         bus_name,
                         G_BUS_NAME_OWNER_FLAGS_NONE,
                         on_name_acquired,
                         on_name_lost,
                         NULL,
                         NULL);

  self->priv->registration_id = g_dbus_connection_register_object (connection,
                                TASKVIEW_MANAGER_OBJECTPATH,
                                g_dbus_interface_info_ref (interface_info),
                                &manager_vtable,
                                NULL,
                                NULL,
                                &error);

  if (error != NULL)
    {
      g_error ("%s", error->message);
      g_error_free (error);
    }

  self->priv->registration_id_subtree = g_dbus_connection_register_subtree (connection,
                                        TASKVIEW_SUBTREE_OBJECTPATH,
                                        &subtree_vtable,
                                        G_DBUS_SUBTREE_FLAGS_NONE,
                                        NULL,
                                        NULL,
                                        &error);

  if (error != NULL)
    {
      g_error ("%s", error->message);
      g_error_free (error);
    }

  // FIXME: ATM we just request the monitor...
  g_bus_watch_name_on_connection (connection,
                                  TASKVIEW_MONITOR_BUS_NAME,
                                  G_BUS_NAME_WATCHER_FLAGS_AUTO_START,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL);
}

/* ----------------------------------------------------------------------------------------------------
 * GObject
 * ---------------------------------------------------------------------------------------------------- */

static GObject*
taskview_manager_constructor (GType type,
                              guint n_construct_params,
                              GObjectConstructParam *construct_params)
{
  static GObject *self = NULL;

  if (self == NULL)
    {
      self = G_OBJECT_CLASS (taskview_manager_parent_class)->constructor (type, n_construct_params, construct_params);
      g_object_add_weak_pointer (self, (gpointer) &self);
      return self;
    }

  return g_object_ref (self);
}

static void
taskview_manager_dispose (GObject *gobject)
{
  TaskviewManager *self = TASKVIEW_MANAGER (gobject);

  if (self->priv->connection)
    {
      g_bus_unown_name (self->priv->owner_id);
      g_dbus_connection_unregister_object (self->priv->connection,
                                           self->priv->registration_id);
      g_dbus_connection_unregister_subtree (self->priv->connection,
                                            self->priv->registration_id_subtree);
      g_object_unref (self->priv->connection);
      self->priv->connection = NULL;
    }

  G_OBJECT_CLASS (taskview_manager_parent_class)->dispose (gobject);
}

static void
taskview_manager_finalize (GObject *gobject)
{
  TaskviewManager *self = TASKVIEW_MANAGER (gobject);

  g_dbus_node_info_unref (introspection_data);

  G_OBJECT_CLASS (taskview_manager_parent_class)->finalize (gobject);
}


static void
taskview_manager_class_init (TaskviewManagerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->constructor = taskview_manager_constructor;
  gobject_class->dispose = taskview_manager_dispose;
  gobject_class->finalize = taskview_manager_finalize;

  g_type_class_add_private (gobject_class, sizeof (TaskviewManagerPrivate));
}

static void
taskview_manager_init (TaskviewManager *self)
{
  self->priv = TASKVIEW_MANAGER_GET_PRIVATE (self);
  self->priv->connection = NULL;

  /* Cache */
  //self->priv->cache = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_hash_table_unref);
  
  /* Build introspection */
  introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
  g_assert (introspection_data != NULL);
  interface_info = g_dbus_node_info_lookup_interface (introspection_data, TASKVIEW_MANAGER_INTERFACE_NAME);
  g_assert (interface_info != NULL);

  g_bus_get (G_BUS_TYPE_SESSION,
             NULL,
             on_bus_acquired,
             self);
}

/* ----------------------------------------------------------------------------------------------------
 * Accessor functions
 * ---------------------------------------------------------------------------------------------------- */

TaskviewManager*
taskview_manager_dup (void)
{
  return g_object_new (TASKVIEW_TYPE_MANAGER, NULL);
}

/* ----------------------------------------------------------------------------------------------------
 * D-Bus Signals
 * ---------------------------------------------------------------------------------------------------- */

void
taskview_manager_tv_changed (TaskviewManager *self,
                             const gchar *object_path,
                             gboolean mode)
{
  g_return_if_fail (TASKVIEW_IS_MANAGER (self));
  g_return_if_fail (g_variant_is_object_path (object_path));

  GError *error = NULL;

  if (self->priv->connection != NULL)
    {
      g_dbus_connection_emit_signal (self->priv->connection,
                                     NULL,
                                     TASKVIEW_MANAGER_OBJECTPATH,
                                     TASKVIEW_MANAGER_INTERFACE_NAME,
                                     mode ? "TaskViewAdded" : "TaskViewRemoved",
                                     g_variant_new ("(o)", object_path),
                                     &error);

      if (error != NULL)
        {
          g_error ("%s", error->message);
          g_error_free (error);
        }
    }
}

void
taskview_manager_property_changed (TaskviewManager *self,
                                   const gchar *object_path,
                                   const gchar *interface,
                                   GVariantBuilder *changed_properties,
                                   GVariantBuilder *invalidated_properties)
{
  g_return_if_fail (TASKVIEW_IS_MANAGER (self));
  g_return_if_fail (g_variant_is_object_path (object_path));

  GError *error = NULL;

  if (self->priv->connection != NULL && changed_properties != NULL)
    {
      g_dbus_connection_emit_signal (self->priv->connection,
                                     NULL,
                                     object_path,
                                     "org.freedesktop.DBus.Properties",
                                     "PropertiesChanged",
                                     g_variant_new ("(sa{sv}as)",
                                                    interface,
                                                    changed_properties,
                                                    invalidated_properties),
                                     &error);
      if (error != NULL)
        {
          g_error ("%s", error->message);
          g_error_free (error);
        }
    }
}

void
clear_cache ()
{

}
